import json
import logging
import pathlib

import parse


from types import SimpleNamespace


logger = logging.getLogger(__name__)


class ItemSpec(SimpleNamespace):

    def template_repr(self, template):
        return template.format(**self.__dict__)

    def json_repr(self):
        return json.dumps(self.__dict__)

    def __hash__(self):
        return self.json_repr().__hash__()


def item_spec_from_fpath(fpath):
    with open(fpath) as fh:
        metadata = json.load(fh)

    return ItemSpec(**metadata["spec"])


def get_all_specs(dirpath):

    dirpath = pathlib.Path(dirpath)
    fpath_iter = dirpath.rglob("*.json")
    specs = [item_spec_from_fpath(fpath) for fpath in fpath_iter]

    return specs


def specs_from_fpath_metadata(dirpath, parse_templates):

    def metadata_from_fname(fname):
        for template in parse_templates:
            result = parse.parse(template, fname)
            if result is not None:
                return result.named

        return None

    fpath_iter = pathlib.Path(dirpath).iterdir()
    fname_iter = (fpath.name for fpath in fpath_iter)

    specs = []
    for fname in fname_iter:
        result = metadata_from_fname(fname)
        if result:
            specs.append(ItemSpec(**result))
        else:
            logger.warning(f"Couldn't parse {fname}")

    return specs



import io

import ruamel.yaml


class Config(object):

    def __init__(self, raw_config):
        self.raw_config = raw_config

    @classmethod
    def from_fpath(cls, yaml_fpath):
        yaml = ruamel.yaml.YAML()
        with open(yaml_fpath) as fh:
            raw_config = yaml.load(fh)

        return cls(raw_config)

    def __getattr__(self, name):
        return self.raw_config[name]

    def as_readme_format(self):
        yaml = ruamel.yaml.YAML()

        readme_dict = {
            "config": self.raw_config
        }

        with io.StringIO() as f:
            yaml.dump(readme_dict, f)
            readme_content = f.getvalue()

        return readme_content



import pathlib

import click

from runtools.spec import get_all_specs


@click.command()
@click.argument("working_dir")
def list_working_dir(working_dir):

    specs = get_all_specs(working_dir)
    reprs = [spec.__repr__() for spec in specs]
    print('\n'.join(sorted(reprs)))



import pathlib
import collections

import dtoolcore

from dtoolbioimage import Image as dbiImage



def iter_identifiers_with_extension(dataset, extension):

    for idn in dataset.identifiers:
        if dataset.item_properties(idn)["relpath"].endswith(extension):
            yield idn


class IndexedDirtree(object):

    def __init__(self, base_dirpath):
        self.base_dirpath = pathlib.Path(base_dirpath)
        self.index = {
            dtoolcore.utils.generate_identifier(str(rpath)): rpath
            for rpath in self.relpath_iter
        }

    @property
    def abspath_iter(self):
        return (
            fpath
            for fpath in self.base_dirpath.rglob("*")
            if fpath.is_file()
        )

    @property
    def relpath_iter(self):
        return (
            fpath.relative_to(self.base_dirpath)
            for fpath in self.abspath_iter
        )

    @property
    def identifiers(self):
        return list(self.index)

    def item_properties(self, idn):
        rpath = self.index[idn]
        return {
            'relpath': str(rpath)
        }

    def item_content_abspath(self, idn):
        return self.base_dirpath/self.index[idn]

    def __len__(self):
        return len(self.identifiers)


class ImageDataSetView(collections.abc.Sequence):

    def __init__(self, container):
        self.container = container
        self.identifiers = list(iter_identifiers_with_extension(container, ".jpg"))

    def __len__(self):
        return len(self.identifiers)

    def __getitem__(self, idx):
        idn = self.identifiers[idx]
        return self.item_by_idn(idn)

    def item_by_idn(self, idn):
        fpath = self.container.item_content_abspath(idn)
        im = dbiImage.from_file(fpath)
        im.properties = self.container.item_properties(idn)
        return im




import io
import sys
import json
import time
import uuid
import logging
import pathlib
import datetime

import ruamel.yaml
import numpy as np


logger = logging.getLogger("runtools")


def one_to_one(input_ds, idns, process_func, output_ds):

    t_start = time.time()
    t_hist = []
    results = []
    for n, idn in enumerate(idns, start=1):
        if len(t_hist):
            t_mean = np.mean(t_hist)
        else:
            t_mean = 1

        n_remaining = len(idns) - n
        t_remaining = t_mean * n_remaining
        logger.info(f"Procesing [{n}/{len(idns)}]: {idn}, {t_remaining:.2f}")

        t_pre = time.time()
        result = process_func(input_ds, idn, output_ds)
        results.append(result)
        t_hist.append(time.time() - t_pre)

    runstats = {
        "total_time": time.time() - t_start,
        "n_items": len(idns)
    }

    return results, runstats


def create_readme(config, runstats):
    yaml = ruamel.yaml.YAML()

    readme_dict = {
        "config": config.raw_config,
        "runstats": runstats
    }

    with io.StringIO() as f:
        yaml.dump(readme_dict, f)
        readme_content = f.getvalue()

    return readme_content


def run_process(spec_iter, measure_func, config):

    n_items = len(spec_iter)

    all_measures = []
    t_start = time.time()
    t_hist = []

    working_dirpath = pathlib.Path(config.working_dirpath)
    working_dirpath.mkdir(exist_ok=True, parents=True)

    for n, spec in enumerate(spec_iter, start=1):
        logger.info(f"Processing {spec}")

        t_pre = time.time()
        measures = measure_func(spec)
        t_hist.append(time.time() - t_pre)

        prefix = uuid.uuid4()
        data_fname = f"{prefix}.csv"

        metadata = {
            "spec": spec.__dict__,
            "data_fname": data_fname
        }
        with open(working_dirpath/f"{prefix}.json", "w") as fh:
            json.dump(metadata, fh, indent=2)
        measures.to_csv(working_dirpath/data_fname)

        time_string = f"Item {n}/{n_items}, {t_hist[-1]:.2f}s, total elapsed: {sum(t_hist):.2f}"

        if n > 0:
            t_mean = np.mean(t_hist)
            t_expected = (n_items -n) * t_mean
            dt_expected = datetime.datetime.now() + datetime.timedelta(0, int(t_expected))
            dt_expected_str = dt_expected.strftime("%H:%M")
            time_string += f" estimate {t_expected:.2f} remaining, ETA {dt_expected_str}"

        logger.info(time_string)

        all_measures.append(measures)

    return all_measures
